searchNodes=[{"doc":"Collection of common optimizers and optimization utilities.","ref":"Polaris.html","title":"Polaris","type":"module"},{"doc":"Implementations of common gradient-based optimization algorithms. All of the methods in this module are written in terms of the update methods defined in Polaris.Updates . Polaris treats optimizers as the tuple: { init_fn , update_fn } where init_fn returns an initial optimizer state and update_fn scales input gradients. init_fn accepts a model's parameters and attaches state to each parameter. update_fn accepts gradients, optimizer state, and current model parameters and returns updated optimizer state and gradients. Custom optimizers are often created via the Polaris.Updates API. Example Consider the following usage of the Adam optimizer in a basic update function (assuming objective and the dataset are defined elsewhere): defmodule Learning do import Nx.Defn defn init ( params , init_fn ) do init_fn . ( params ) end defn update ( params , optimizer_state , inputs , targets , update_fn ) do { loss , gradient } = value_and_grad ( params , &amp; objective ( &amp;1 , inputs , targets ) ) { scaled_updates , new_optimizer_state } = update_fn . ( gradient , optimizer_state , params ) { Polaris.Updates . apply_updates ( params , scaled_updates ) , new_optimizer_state , loss } end end { model_params , _key } = Nx.Random . uniform ( key , shape : { 784 , 10 } ) { init_fn , update_fn } = Polaris.Optimizers . adam ( 0.005 ) optimizer_state = Learning . init ( params , init_fn ) { new_params , new_optimizer_state , loss } = Learning . update ( params , optimizer_state , inputs , targets , update_fn )","ref":"Polaris.Optimizers.html","title":"Polaris.Optimizers","type":"module"},{"doc":"Adabelief optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-3 :b1 - first moment decay. Defaults to 0.9 :b2 - second moment decay. Defaults to 0.999 :eps - numerical stability term. Defaults to 0.0 :eps_root - numerical stability term. Defaults to 1.0e-16 References AdaBelief Optimizer: Adapting Stepsizes by the Belief in Observed Gradients","ref":"Polaris.Optimizers.html#adabelief/1","title":"Polaris.Optimizers.adabelief/1","type":"function"},{"doc":"Adagrad optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-3 :eps - numerical stability term. Defaults to 1.0e-7 References Adaptive Subgradient Methods for Online Learning and Stochastic Optimization","ref":"Polaris.Optimizers.html#adagrad/1","title":"Polaris.Optimizers.adagrad/1","type":"function"},{"doc":"Adam optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-3 :b1 - first moment decay. Defaults to 0.9 :b2 - second moment decay. Defaults to 0.999 :eps - numerical stability term. Defaults to 1.0e-8 :eps_root - numerical stability term. Defaults to 1.0e-15 References Adam: A Method for Stochastic Optimization","ref":"Polaris.Optimizers.html#adam/1","title":"Polaris.Optimizers.adam/1","type":"function"},{"doc":"Adam with weight decay optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-3 :b1 - first moment decay. Defaults to 0.9 :b2 - second moment decay. Defaults to 0.999 :eps - numerical stability term. Defaults to 1.0e-8 :eps_root - numerical stability term. Defaults to 0.0 :decay - weight decay. Defaults to 0.0","ref":"Polaris.Optimizers.html#adamw/1","title":"Polaris.Optimizers.adamw/1","type":"function"},{"doc":"Lamb optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-2 :b1 - first moment decay. Defaults to 0.9 :b2 - second moment decay. Defaults to 0.999 :eps - numerical stability term. Defaults to 1.0e-8 :eps_root - numerical stability term. Defaults to 0.0 :decay - weight decay. Defaults to 0.0 :min_norm - minimum norm value. Defaults to 0.0 References Large Batch Optimization for Deep Learning: Training BERT in 76 minutes","ref":"Polaris.Optimizers.html#lamb/1","title":"Polaris.Optimizers.lamb/1","type":"function"},{"doc":"Noisy SGD optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-2 :eta - used to compute variance of noise distribution. Defaults to 0.1 :gamma - used to compute variance of noise distribution. Defaults to 0.55","ref":"Polaris.Optimizers.html#noisy_sgd/1","title":"Polaris.Optimizers.noisy_sgd/1","type":"function"},{"doc":"Rectified Adam optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-3 :b1 - first moment decay. Defaults to 0.9 :b2 - second moment decay. Defaults to 0.999 :eps - numerical stability term. Defaults to 1.0e-8 :eps_root - numerical stability term. Defaults to 0.0 :threshold - threshold term. Defaults to 5.0 References On the Variance of Adaptive Learning Rate and Beyond","ref":"Polaris.Optimizers.html#radam/1","title":"Polaris.Optimizers.radam/1","type":"function"},{"doc":"RMSProp optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-2 :centered - whether to scale by centered root of EMA of squares. Defaults to false :momentum - momentum term. If set, uses SGD with momentum and decay set to value of this term. :nesterov - whether or not to use nesterov momentum. Defaults to false :initial_scale - initial value of EMA. Defaults to 0.0 :decay - EMA decay rate. Defaults to 0.9 :eps - numerical stability term. Defaults to 1.0e-8","ref":"Polaris.Optimizers.html#rmsprop/1","title":"Polaris.Optimizers.rmsprop/1","type":"function"},{"doc":"SGD optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-2 :momentum - momentum term. If set, uses SGD with momentum and decay set to value of this term. :nesterov - whether or not to use nesterov momentum. Defaults to false","ref":"Polaris.Optimizers.html#sgd/1","title":"Polaris.Optimizers.sgd/1","type":"function"},{"doc":"Yogi optimizer. Options :learning_rate - the learning rate for the optimizer. Defaults to 1.0e-2 :initial_accumulator_value - initial value for first and second moment. Defaults to 0.0 :b1 - first moment decay. Defaults to 0.9 :b2 - second moment decay. Defaults to 0.999 :eps - numerical stability term. Defaults to 1.0e-8 :eps_root - numerical stability term. Defaults to 0.0 References Adaptive Methods for Nonconvex Optimization","ref":"Polaris.Optimizers.html#yogi/1","title":"Polaris.Optimizers.yogi/1","type":"function"},{"doc":"Parameter Schedules. Parameter schedules are often used to anneal hyperparameters such as the learning rate during the training process. Schedules provide a mapping from the current time step to a learning rate or another hyperparameter. Choosing a good learning rate and consequently a good learning rate schedule is typically a process of trial and error. Learning rates should be relatively small such that the learning curve does not oscillate violently during the training process, but not so small that learning proceeds too slowly. Using a schedule slowly decreases oscillations during the training process such that, as the model converges, training also becomes more stable. All of the functions in this module are implemented as numerical functions and can be JIT or AOT compiled with any supported Nx compiler.","ref":"Polaris.Schedules.html","title":"Polaris.Schedules","type":"module"},{"doc":"Constant schedule. $$\\gamma(t) = \\gamma_0$$","ref":"Polaris.Schedules.html#constant/2","title":"Polaris.Schedules.constant/2","type":"function"},{"doc":"Cosine decay schedule. $$\\gamma(t) = \\gamma_0 * \\left(\\frac{1}{2}(1 - \\alpha)(1 + \\cos\\pi \\frac{t}{k}) + \\alpha\\right)$$ Options :decay_steps - number of steps to apply decay for. $k$ in above formulation. Defaults to 10 :alpha - minimum value of multiplier adjusting learning rate. $\\alpha$ in above formulation. Defaults to 0.0 References SGDR: Stochastic Gradient Descent with Warm Restarts","ref":"Polaris.Schedules.html#cosine_decay/2","title":"Polaris.Schedules.cosine_decay/2","type":"function"},{"doc":"Exponential decay schedule. $$\\gamma(t) = \\gamma_0 * r^{\\frac{t}{k}}$$ Options :decay_rate - rate of decay. $r$ in above formulation. Defaults to 0.95 :transition_steps - steps per transition. $k$ in above formulation. Defaults to 10 :transition_begin - step to begin transition. Defaults to 0 :staircase - discretize outputs. Defaults to false","ref":"Polaris.Schedules.html#exponential_decay/2","title":"Polaris.Schedules.exponential_decay/2","type":"function"},{"doc":"Linear decay schedule. Options :warmup - scheduler warmup steps. Defaults to 0 :steps - total number of decay steps. Defaults to 1000","ref":"Polaris.Schedules.html#linear_decay/2","title":"Polaris.Schedules.linear_decay/2","type":"function"},{"doc":"Polynomial schedule. $$\\gamma(t) = (\\gamma_0 - \\gamma_n) * (1 - \\frac{t}{k})^p$$ Options :end_value - end value of annealed scalar. $\\gamma_n$ in above formulation. Defaults to 1.0e-3 :power - power of polynomial. $p$ in above formulation. Defaults to 2 :transition_steps - number of steps over which annealing takes place. $k$ in above formulation. Defaults to 10","ref":"Polaris.Schedules.html#polynomial_decay/2","title":"Polaris.Schedules.polynomial_decay/2","type":"function"},{"doc":"Parameter update methods. Update methods transform the input tensor in some way, usually by scaling or shifting the input with respect to some input state. Update methods are composed to create more advanced optimization methods such as AdaGrad or Adam. Each update returns a tuple: { init_fn , update_fn } Which represent a state initialization and state update function respectively. While each method in the Updates API is a regular Elixir function, the two methods they return are implemented as defn , so they can be accelerated using any Nx backend or compiler. Update methods are just combinators that can be arbitrarily composed to create complex optimizers. For example, the Adam optimizer in Polaris.Optimizers is implemented as: def adam ( opts \\\\ [ ] ) do opts |&gt; Polaris.Updates . scale_by_adam ( ) |&gt; Polaris.Updates . scale ( - opts [ :learning_rate ] ) end Updates are maps of updates, often associated with parameters of the same names. Using Polaris.Updates.apply_updates/3 will merge updates and parameters by adding associated parameters and updates, and ensuring any given model state is preserved. Custom combinators You can create your own combinators using the stateless/2 and stateful/3 primitives. Every update method in this module is implemented in terms of one of these two primitives. stateless/2 represents a stateless update: def scale ( combinator \\\\ Polaris.Updates . identity ( ) , step_size ) do stateless ( combinator , &amp; apply_scale ( &amp;1 , &amp;2 , step_size ) ) end defnp apply_scale ( x , _params , step ) do deep_new ( updates , fn x -&gt; Nx . multiply ( x , step ) end ) end Notice how the function given to stateless/2 is defined within defn . This is what allows the anonymous functions returned by Polaris.Updates to be used inside defn . stateful/3 represents a stateful update and follows the same pattern: def my_stateful_update ( updates ) do Polaris.Updates . stateful ( updates , &amp; init_my_update / 1 , &amp; apply_my_update / 2 ) end defnp init_my_update ( params ) do state = zeros_like ( params , type : :f32 ) %{ state : state } end defnp apply_my_update ( updates , state ) do new_state = deep_new ( state , fn v -&gt; Nx . add ( v , 0.01 ) end ) updates = deep_merge ( updates , state , fn g , z -&gt; Nx . multiply ( g , z ) end ) { updates , %{ state : new_state } } end State associated with individual parameters should have keys that match the keys of the parameter. For example, if you have parameters %{kernel: kernel} with associated states mu and nu representing the first and second moments, your state should look something like: %{ mu : %{ kernel : kernel_mu } nu : %{ kernel : kernel_nu } }","ref":"Polaris.Updates.html","title":"Polaris.Updates","type":"module"},{"doc":"Adds decayed weights to updates. Commonly used as a regularization strategy. Options * ` :decay ` - Rate of decay . Defaults to ` 0.0 ` .","ref":"Polaris.Updates.html#add_decayed_weights/1","title":"Polaris.Updates.add_decayed_weights/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#add_decayed_weights/2","title":"Polaris.Updates.add_decayed_weights/2","type":"function"},{"doc":"Adds random Gaussian noise to the input. Options * ` :seed ` - Random seed to use . Defaults to the current system time . * ` :eta ` - Controls amount of noise to add . Defaults to ` 0.01 ` . * ` :gamma ` - Controls amount of noise to add . Defaults to ` 0.55 ` .","ref":"Polaris.Updates.html#add_noise/1","title":"Polaris.Updates.add_noise/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#add_noise/2","title":"Polaris.Updates.add_noise/2","type":"function"},{"doc":"Applies updates to params and updates state parameters with given state map.","ref":"Polaris.Updates.html#apply_updates/3","title":"Polaris.Updates.apply_updates/3","type":"function"},{"doc":"Centralizes input by shifting updates by their mean.","ref":"Polaris.Updates.html#centralize/1","title":"Polaris.Updates.centralize/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#centralize/2","title":"Polaris.Updates.centralize/2","type":"function"},{"doc":"Clips input between -delta and delta. Options :delta - maximum absolute value of the input. Defaults to 2.0","ref":"Polaris.Updates.html#clip/1","title":"Polaris.Updates.clip/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#clip/2","title":"Polaris.Updates.clip/2","type":"function"},{"doc":"Clips input using input global norm. Options :max_norm - maximum norm value of input. Defaults to 1.0","ref":"Polaris.Updates.html#clip_by_global_norm/1","title":"Polaris.Updates.clip_by_global_norm/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#clip_by_global_norm/2","title":"Polaris.Updates.clip_by_global_norm/2","type":"function"},{"doc":"Composes two updates. This is useful for extending optimizers without having to reimplement them. For example, you can implement gradient centralization: import Polaris.Updates Polaris.Updates . compose ( Polaris.Updates . centralize ( ) , Polaris.Optimizers . rmsprop ( ) ) This is equivalent to: Polaris.Updates . centralize ( ) |&gt; Polaris.Updates . scale_by_rms ( )","ref":"Polaris.Updates.html#compose/2","title":"Polaris.Updates.compose/2","type":"function"},{"doc":"Returns the identity update. This is often as the initial update in many functions in this module.","ref":"Polaris.Updates.html#identity/0","title":"Polaris.Updates.identity/0","type":"function"},{"doc":"","ref":"Polaris.Updates.html#identity/1","title":"Polaris.Updates.identity/1","type":"function"},{"doc":"Scales input by a fixed step size. $$f(x_i) = \\alpha x_i$$","ref":"Polaris.Updates.html#scale/2","title":"Polaris.Updates.scale/2","type":"function"},{"doc":"Scales input according to Adam algorithm. Options * ` :b1 ` - first moment decay . Defaults to ` 0.9 ` * ` :b2 ` - second moment decay . Defaults to ` 0.999 ` * ` :eps ` - numerical stability term . Defaults to ` 1.0e-8 ` * ` :eps_root ` - numerical stability term . Defaults to ` 1.0e-15 ` References Adam: A Method for Stochastic Optimization","ref":"Polaris.Updates.html#scale_by_adam/1","title":"Polaris.Updates.scale_by_adam/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_adam/2","title":"Polaris.Updates.scale_by_adam/2","type":"function"},{"doc":"Scales input according to the AdaBelief algorithm. Options * ` :b1 ` - first moment decay . Defaults to ` 0.9 ` . * ` :b2 ` - second moment decay . Defaults to ` 0.999 ` . * ` :eps ` - numerical stability term . Defaults to ` 0.0 ` . * ` :eps_root ` - numerical stability term . Defaults to ` 1.0e-16 ` . References AdaBelief Optimizer: Adapting Stepsizes by the Belief in Observed Gradients","ref":"Polaris.Updates.html#scale_by_belief/1","title":"Polaris.Updates.scale_by_belief/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_belief/2","title":"Polaris.Updates.scale_by_belief/2","type":"function"},{"doc":"Scale input according to the Rectified Adam algorithm. Options * ` :b1 ` - first moment decay . Defaults to ` 0.9 ` * ` :b2 ` - second moment decay . Defaults to ` 0.999 ` * ` :eps ` - numerical stability term . Defaults to ` 1.0e-8 ` * ` :eps_root ` - numerical stability term . Defaults to ` 0.0 ` * ` :threshold ` - threshold for variance . Defaults to ` 5.0 ` References On the Variance of the Adaptive Learning Rate and Beyond","ref":"Polaris.Updates.html#scale_by_radam/1","title":"Polaris.Updates.scale_by_radam/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_radam/2","title":"Polaris.Updates.scale_by_radam/2","type":"function"},{"doc":"Scales input by the root of the EMA of squared inputs. Options * ` :decay ` - EMA decay rate . Defaults to ` 0.9 ` . * ` :eps ` - numerical stability term . Defaults to ` 1.0e-8 ` . References Overview of mini-batch gradient descent","ref":"Polaris.Updates.html#scale_by_rms/1","title":"Polaris.Updates.scale_by_rms/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_rms/2","title":"Polaris.Updates.scale_by_rms/2","type":"function"},{"doc":"Scales input by the root of all prior squared inputs. Options * ` :eps ` - numerical stability term . Defaults to ` 1.0e-7 `","ref":"Polaris.Updates.html#scale_by_rss/1","title":"Polaris.Updates.scale_by_rss/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_rss/2","title":"Polaris.Updates.scale_by_rss/2","type":"function"},{"doc":"Scales input using the given schedule function. This can be useful for implementing learning rate schedules. The number of update iterations is tracked by an internal counter. You might need to update the schedule to operate on per-batch schedule rather than per-epoch.","ref":"Polaris.Updates.html#scale_by_schedule/2","title":"Polaris.Updates.scale_by_schedule/2","type":"function"},{"doc":"Scales input by a tunable learning rate which can be manipulated by external APIs such as Polaris's Loop API. $$f(x_i) = \\alpha x_i$$","ref":"Polaris.Updates.html#scale_by_state/1","title":"Polaris.Updates.scale_by_state/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_state/2","title":"Polaris.Updates.scale_by_state/2","type":"function"},{"doc":"Scales input by the root of the centered EMA of squared inputs. Options * ` :decay ` - EMA decay rate . Defaults to ` 0.9 ` . * ` :eps ` - numerical stability term . Defaults to ` 1.0e-8 ` . References Overview of mini-batch gradient descent","ref":"Polaris.Updates.html#scale_by_stddev/1","title":"Polaris.Updates.scale_by_stddev/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_stddev/2","title":"Polaris.Updates.scale_by_stddev/2","type":"function"},{"doc":"Scale by trust ratio. Options * ` :min_norm ` - Min norm to clip . Defaults to ` 0.0 ` . * ` :trust_coefficient ` - Trust coefficient . Defaults to ` 1.0 ` . * ` :eps ` - Numerical stability term . Defaults to ` 0.0 ` .","ref":"Polaris.Updates.html#scale_by_trust_ratio/1","title":"Polaris.Updates.scale_by_trust_ratio/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_trust_ratio/2","title":"Polaris.Updates.scale_by_trust_ratio/2","type":"function"},{"doc":"Scale input according to the Yogi algorithm. Options * ` :initial_accumulator_value ` - Initial state accumulator value . * ` :b1 ` - first moment decay . Defaults to ` 0.9 ` * ` :b2 ` - second moment decay . Defaults to ` 0.999 ` * ` :eps ` - numerical stability term . Defaults to ` 1.0e-8 ` * ` :eps_root ` - numerical stability term . Defaults to ` 0.0 ` References * [ Adaptive Methods for Nonconvex Optimization ] ( https :// proceedings . neurips . cc / paper / 2018 / file / 90365351 ccc7437a1309dc64e4db32a3 - Paper . pdf )","ref":"Polaris.Updates.html#scale_by_yogi/1","title":"Polaris.Updates.scale_by_yogi/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#scale_by_yogi/2","title":"Polaris.Updates.scale_by_yogi/2","type":"function"},{"doc":"Represents a stateful update. Stateful updates require some update state, such as momentum or RMS of previous updates. Therefore you must implement some initialization function as well as an update function.","ref":"Polaris.Updates.html#stateful/3","title":"Polaris.Updates.stateful/3","type":"function"},{"doc":"Represents a stateless update. Stateless updates do not depend on an update state and thus only require an implementation of an update function.","ref":"Polaris.Updates.html#stateless/2","title":"Polaris.Updates.stateless/2","type":"function"},{"doc":"Trace inputs with past inputs. Options :decay - decay rate for tracing past updates. Defaults to 0.9 :nesterov - whether to use Nesterov momentum. Defaults to false","ref":"Polaris.Updates.html#trace/1","title":"Polaris.Updates.trace/1","type":"function"},{"doc":"","ref":"Polaris.Updates.html#trace/2","title":"Polaris.Updates.trace/2","type":"function"}]